use strict;
use warnings;

package CORBA::JAVA;

our $VERSION = '2.66';

use CORBA::JAVA::ClassVisitor;
use CORBA::JAVA::LiteralVisitor;
use CORBA::JAVA::NameVisitor;
use CORBA::JAVA::Name2Visitor;
use CORBA::JAVA::UidVisitor;

1;

